```{gql:directive} {{ signature }}

{% if description %}
{{ description }}

{% endif %}
{% for arg, arg_description in args.items() %}
:argument {{ arg }}: {% if arg_description %}{{ arg_description|indent(3) }}{% endif +%}
{% endfor %}
```